%SUPERMARKET

% Check if object is graspable
is_graspable([W,Z,Y,X],Res):-
	print('W : '), print(W), nl,
	print('Z : '), print(Z), nl,
	print('Y : '), print(Y), nl,
	print('X : '), print(X), nl,
    P is atan2(2*Y*Z + 2*W*X, W*W - X*X - Y*Y + Z*Z), 
    R is atan2(2*X*Y + 2*W*Z, W*W + X*X - Y*Y - Z*Z),
    Prea is (2*W*Y - 2*X*Z),
	%print('Prea : '), print(Prea), nl,
    (Prea > 1 ->
       A is asin(1)
    | otherwise ->
        (Prea < -1 ->
            A is asin(-1)
        | otherwise ->
            A is asin(2*W*Y - 2*X*Z)
        )
    ),
	print('Roll : '), print(R), nl,
	print('Pitch: '), print(P), nl,
	print('Yaw  : '), print(A), nl,
    (abs(R) > 0.7854 ->
       Res = not_graspable
    | otherwise ->
       (abs(P) > 0.7854 ->
          Res = not_graspable
       | otherwise ->
          Res = graspable
       ) %ignoring yaw for now
    ),
	print('Result : '), print(Res), nl.
	%assign_func_value(Res).
    

%Parser

supermarket_parser('do you have a dairy product that is on discount'):-
    assign_func_value(bring_me(dairy,on_discount=>yes)).
    
supermarket_parser('do you have toilet paper'):-
    assign_func_value(bring_me('toilet paper')).

supermarket_parser('do you have beer'):-
    assign_func_value(bring_me(beer)).

supermarket_parser('do you have a drink for all ages'):-
    assign_func_value(bring_me(drink,age=>all)).
    
supermarket_parser('do you have cereal'):-
    assign_func_value(bring_me(cereal)).
    
supermarket_parser('no'):-
    assign_func_value(dispatch).


%Existence

check_restriction(existence,RestRestrictions,Input):-
    consult_existence_of_an_object(Input,there_is(Object)),
    open_kb(KB),
    object_property_value(Object,brand,KB,Brand),
    change_object_name(Object,Brand,KB,NewKB),
    save_kb(NewKB),
    assign_func_value(embedded_sit(say(['Yes.',Brand],Status),empty,verify_restriction(RestRestrictions,Brand))).

check_restriction(existence,RestRestrictions,Input):-
    consult_existence_of_an_object(Input,unknown_object(Object)),
    assign_func_value(embedded_sit(say(['I dont have information in my knowledge base about that product'],Status),empty,success)).
    
check_object_is_from_class([],Value,Class,KB,[]).

check_object_is_from_class([Object:Value|T],Value,Class,KB,Object):-
    objects_of_a_class(Class,KB,AllObjects),
    isElement(Object,AllObjects).
    
check_object_is_from_class([Object:_|T],Value,Class,KB,Object):-
    check_object_is_from_class(T,Value,Class,KB,Object).
    
consult_existence_of_an_object(bring_me(Class,Property=>Value),there_is(Object)):-
    open_kb(KB),
    property_extension(Property,KB,List),
    check_object_is_from_class(List,Value,Class,KB,Object).

consult_existence_of_an_object(bring_me(Class),there_is(Object)):-
    open_kb(KB),
    there_is_class(Class,KB,yes),
    objects_of_a_class(Class,KB,[Object|Rest]).
    
consult_existence_of_an_object(bring_me(Class),unknown_object(Object)):-
    open_kb(KB),
    there_is_class(Class,KB,unknown).
    

%Disponibility

check_restriction(disponibility,RestRestrictions,Input):-
    check_inventory_of_an_object(Input,EmbeddedModel,RestRestrictions,Input),
    assign_func_value(EmbeddedModel).

check_inventory_of_an_object(Object,EmbeddedModel,RestRestrictions,Input):-
    open_kb(KB),
    object_property_value(Object,inv,KB,Number),
    verify_value_in_inventory(Number,Object,EmbeddedModel,RestRestrictions,Input).

verify_value_in_inventory(Number,Object,Embedded,RestRestrictions,Input):-
    ( Number == 0 ->
      Embedded=embedded_sit(say(['But we dont have',Object,'in inventory'],Status),empty,success),
      register_failed_order(Object,not_disponible)
     | otherwise ->
       Embedded = embedded_sit(say('.',Status),verify_value_in_inventory(Number,Object,empty),verify_restriction(RestRestrictions,Input))
    ).

register_failed_order(Object,Reason):-
    open_kb(KB),
    add_object(bring(Object),failed_orders,KB,NewKB),
    add_object_property(bring(Object),reason,Reason,NewKB,NewKB2),
    save_kb(NewKB2).


%Graspability

check_restriction(graspability,RestRestrictions,Input):-
    check_object_graspable(Input,Value),
    object_graspable_next_sit(Value,NextSit,RestRestrictions,Input),
    assign_func_value(NextSit).

check_object_graspable(Object,Value):-
    open_kb(KB),
    object_property_value(Object,graspable,KB,Value).
    
object_graspable_next_sit(yes,NextSit,RestRestrictions,Input):-
    NextSit = embedded_sit(say('.',Status),empty,verify_restriction(RestRestrictions,Input)).

object_graspable_next_sit(no,NextSit,RestRestrictions,Input):-
    NextSit = embedded_sit(say('But I can not grasp it',Status),empty,success).
    

%Age restriction

check_restriction(age_restriction,RestRestrictions,Input):-
    check_object_age_restriction(Input,Value),
    age_restriction_next_sit(Value,NextSit,RestRestrictions,Input),
    assign_func_value(NextSit).

check_object_age_restriction(Object,Value):-
    open_kb(KB),
    object_property_value(Object,age,KB,Value).
    
age_restriction_next_sit(all,NextSit,RestRestrictions,Input):-
    NextSit = embedded_sit(say('I will bring it to you',Status),empty,success),
    register_order(Input).

age_restriction_next_sit(18,NextSit,RestRestrictions,Object):-
    NextSit = embedded_sit(ask('Are you over eighteen',yesno,false,1,Answer,Status),empty,verify_restriction(RestRestrictions,[Object,Answer])).


%Age of client

check_restriction(age_client,RestRestrictions,[Object,Answer]):-
    age_client_next_sit(Answer,NextSit,RestRestrictions,Object),
    assign_func_value(NextSit).

age_client_next_sit(yes,NextSit,RestRestrictions,Input):-
    NextSit = embedded_sit(say('I will bring it to you',Status),empty,success),
    register_order(Input).

age_client_next_sit(no,NextSit,RestRestrictions,Input):-
    NextSit = embedded_sit(say(['I cant give you',Input,'because you are very young'],Status),empty,verify_restriction(RestRestrictions,Input)).

%Shelves

get_shelf_of_an_object(Object):-
    open_kb(KB),
    object_property_value(Object,shelf,KB,Shelf),
    assign_func_value(Shelf).


%Register order

register_order(Object):-
    open_kb(KB),
    add_object(bring(Object),orders,KB,NewKB),
    save_kb(NewKB).


%Other restriction
    
check_restriction(FirstRestriction,RestRestrictions,Input):-
    assign_func_value(embedded_sit(say('Checking new restriction',Status),say('Ready'),verify_restriction(RestRestrictions,Input))).
   
%Remove order

remove_order(Object):-
    open_kb(KB),
    rm_object(bring(Object),KB,NewKB),
    save_kb(NewKB),
    assign_func_value(empty).


%Get list of orders

get_list_of_orders:-
    open_kb(KB),
    objects_of_a_class(orders,KB,Orders),    
    assign_func_value(Orders).

get_list_of_failed_orders:-
    open_kb(KB),
    objects_of_a_class(failed_orders,KB,Orders),
    check_failed_order(Orders,[],PossibleOrders),
    assign_func_value(PossibleOrders).

check_failed_order([],[],PossibleOrders):-
    PossibleOrders = [].

check_failed_order([],PossibleOrdersLast,PossibleOrders):-
    PossibleOrders = PossibleOrdersLast.

check_failed_order([bring(Object)|RestOrders],PossibleOrdersLast,PossibleOrders):-
    open_kb(KB),
    object_property_value(bring(Object),reason,KB,Reason),
    ( Reason == not_disponible ->
      object_property_value(Object,inv,KB,Inventory),
      ( Inventory == 1 ->
        append(PossibleOrdersLast,[bring(Object)],PossibleOrdersNext)
        | otherwise ->
        PossibleOrdersNext = PossibleOrdersLast
      )
      | otherwise ->
      PossibleOrdersNext = PossibleOrdersLast
    ),
    check_failed_order(RestOrders,PossibleOrdersNext,PossibleOrders).

%Actualiza supermarket KB

update_object_inv_in_KB(Object,Inventory):-
	print('Updating inventory for : '), print(Object), nl,
    open_kb(KB),
    class_of_an_object(Object,KB,Class),
    change_value_class_property(Class,inv,Inventory,KB,NewKB),
    save_kb(NewKB),
    assign_func_value(empty).
    
get_object_by_property(Property,Value,Object):-
    open_kb(KB),
	print('1 : '), print(Property), print(Value), nl,
    property_extension(Property,KB,List),
	print('2 : '), print(List), nl,
    check_list_by_property(Value,List,Object).

check_list_by_property(Value,[Name:TestValue|Rest],Object):-
	print('3 : '), print(Name), print(TestValue), nl,
    ( Value == TestValue ->
      Object = Name
    | otherwise ->
      check_list_by_property(Value,Rest,Object)
    ).

check_list_by_property(Value,[],Object):-
    Object = empty.

update_object_graspable_in_KB(Object,[Q1,Q2,Q3,Q4]):-
	print('Updating graspable for : '), print(Object), nl,
	is_graspable([Q1,Q2,Q3,Q4],ResGrasp),
    ( ResGrasp == not_graspable ->
         get_object_by_property(brand,Object,ObjectName),
         open_kb(KB),
         change_value_object_property(ObjectName,graspable,no,KB,NewKB),
         save_kb(NewKB)
    | otherwise ->
         print('It is graspable.'), nl
    ),
    assign_func_value(empty).
    
actualize_supermarket_kb:-
    open_kb(KB),
    change_value_class_property(dairy,inv,0,KB,NewKB),
    add_object_property(cereal,graspable,no,NewKB,NewKB2),    
    save_kb(NewKB2),
    assign_func_value(empty).  
   


